//
//  NRosyField.h
//  NRosy
//
//  Created by Daniele Panozzo on 9/6/12.
//  Copyright (c) 2012 Daniele Panozzo. All rights reserved.
//

#ifndef __NRosy__NRosyField__
#define __NRosy__NRosyField__

#include <iostream>
#include <Eigen/Core>
#include <Eigen/Sparse>
#include <vector>

#define USECOMISO

class NRosyField{
public:
  // Init
  NRosyField(const Eigen::MatrixXd& _V, const Eigen::MatrixXi& _F);
  
  // Generate the N-rosy field
  // N degree of the rosy field
  // roundseparately: round the integer variables one at a time, slower but higher quality
  void solve(const int N = 4);
  
  // Set a hard constraint on fid
  // fid: face id
  // v: direction to fix (in 3d)
  void setConstraintHard(const int fid, const Eigen::Vector3d& v);
  
  // Set a soft constraint on fid
  // fid: face id
  // w: weight of the soft constraint, clipped between 0 and 1
  // v: direction to fix (in 3d)
  void setConstraintSoft(const int fid, const double w, const Eigen::Vector3d& v);
  
  // Set the ratio between smoothness and soft constraints (0 -> smoothness only, 1 -> soft constr only)
  void setSoftAlpha(double alpha);
  
  // Reset constraints (at least one constraint must be present or solve will fail)
  void resetConstraints();
  
  // Return the current field
  Eigen::MatrixXd getFieldPerFace();

  // Return the current field (in Ahish's ffield format)
  Eigen::MatrixXd getFFieldPerFace();

  // Compute singularity indexes
  void findCones(int N);

  // Return the singularities
  Eigen::VectorXd getSingularityIndexPerVertex();
  
private:
  
  // Compute angle differences between reference frames
  void computek();
  
  // Remove useless matchings
  void reduceSpace();

  // Prepare the system matrix
  void prepareSystemMatrix(const int N);
  
  // Solve without roundings
  void solveNoRoundings();

#ifdef USECOMISO
  // Solve with roundings using CoMIso
  void solveRoundings();
#endif

  // Round all p and fix
  void roundAndFix();
  
  // Convert a vector in 3d to an angle wrt the local reference system
  double convert3DtoLocal(unsigned fid, const Eigen::Vector3d& v);

  // Convert an angle wrt the local reference system to a 3d vector
  Eigen::Vector3d convertLocalto3D(unsigned fid, double a);

  // Compute the per vertex angle defect 
  Eigen::VectorXd angleDefect();
  
  // Temporary variable for the field
  Eigen::VectorXd angles;
  
  // Hard constraints
  Eigen::VectorXd hard;
  std::vector<bool> isHard;

  // Soft constraints
  Eigen::VectorXd soft;
  Eigen::VectorXd wSoft;
  double          softAlpha;

  // Face Topology
  Eigen::MatrixXi TT, TTi;
  
  // Edge Topology
  Eigen::MatrixXi EV, FE, EF;
  std::vector<bool> isBorderEdge;
  
  // Per Edge information
  // Angle between two reference frames
  Eigen::VectorXd k;
  
  // Jumps
  Eigen::VectorXi p;
  std::vector<bool> pFixed;
  
  // Mesh
  Eigen::MatrixXd V;
  Eigen::MatrixXi F;
  
  // Normals per face
  Eigen::MatrixXd N;
  
  // Singularity index
  Eigen::VectorXd singularityIndex;

  // Reference frame per triangle
  std::vector<Eigen::MatrixXd> TPs;
  
  // System stuff
  Eigen::SparseMatrix<double> A;
  Eigen::VectorXd b;
  Eigen::VectorXi tag_t;
  Eigen::VectorXi tag_p;
  
};



#endif /* defined(__NRosy__NRosyField__) */
